/*************************************************************************** 
 * File        : list.h
 * Date        : 2007/05/27 
 * Author      : Sun Joon Gyu
 * Modified    : 2007/05/27
 * Description : Use this functions to manage and initialize list structure 
 * 		 used in ElfOS Kernel.
 * 		 Most of function listed in below is from linux kernel
 * ***************************************************************************/
#ifndef _LIST_H_
#define _LIST_H_


struct list_head {
	struct list_head *next, *prev;
};

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)

static inline void INIT_LIST_HEAD(struct list_head *list)
{
	list->next = list;
	list->prev = list;
}



static inline void __list_add(struct list_head *new,
			      struct list_head *prev,
			      struct list_head *next)
{
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}



static inline void ListAdd(struct list_head *new, struct list_head *head)
{
	__list_add(new, head, head->next);
}



static inline void ListAddTail(struct list_head *new, struct list_head *head)
{
	__list_add(new, head->prev, head);
}



static inline void __list_del(struct list_head * prev, struct list_head * next)
{
	next->prev = prev;
	prev->next = next;
}



static inline void ListDel(struct list_head *entry)
{
	__list_del(entry->prev, entry->next);
	entry->next = entry;
	entry->prev = entry;
}



static inline int ListEmpty(const struct list_head *head)
{
	return head->next == head;
}




#define ListEntry(ptr, type, member) \
	container_of(ptr, type, member)



#define ListForEach(pos, head) \
	for (pos = (head)->next; prefetch(pos->next), pos != (head); \
        	pos = pos->next)



#define ListForEachEntry(pos, head, member)				\
	for (pos = ListEntry((head)->next, typeof(*pos), member);	\
	     /*prefetch(pos->member.next),*/ &pos->member != (head); 	\
	     pos = ListEntry(pos->member.next, typeof(*pos), member))


/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
        const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
        (type *)( (char *)__mptr - offsetof(type,member) );})


#undef offsetof
#ifdef __compiler_offsetof
#define offsetof(TYPE,MEMBER) __compiler_offsetof(TYPE,MEMBER)
#else
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif


#endif // _LIST_H_
