#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <stdarg.h>

#define TRUE            1
#define FALSE           0

#define MM_MAX_INT      30
#define MM_MAX_TASK     64


#define DEBUG   	0
void DebugPrint(int lv, const char *format, ...)
{
  va_list p_arg;

  if (DEBUG < lv) {
    va_start(p_arg, format);
    vprintf(format, p_arg);
    va_end(p_arg);
  }
}

extern void Kern_TickISR();
//extern void scheduler();

pthread_t mm_mainTID, mm_taskTID[MM_MAX_TASK];
pthread_t mm_nextTID, mm_currTID = -1;
int mm_isrNo = 0;
int mm_interruptVector[MM_MAX_INT];
void (*mm_interruptRoutine[MM_MAX_INT])();
sigset_t mm_interruptMask, mm_blockMask;
int mm_TASKSCHE, mm_TASKSTOP, mm_TASKCONT;
int mm_criticalSectionFlag = FALSE;

void mm_Init();
int mm_ISRInit(int, int, void (*)());
void mm_InterruptInit();
void mm_CriticalSectionEnter();
void mm_CriticalSectionExit();
void mm_TaskScheHandler();
void mm_TaskStopHandler();
void mm_TaskContHandler();
void mm_ContextSwitch(int);
void *mm_TaskPrepare();
void mm_TaskGenerate(int, void *, void *);
void mm_TimerStart();


void mm_Init()
{
  mm_mainTID = pthread_self();

  mm_ISRInit(0, SIGALRM, Kern_TickISR);
//  mm_ISRInit(0, SIGALRM, scheduler);
  mm_ISRInit(1, SIGTERM, mm_TaskScheHandler);
  mm_ISRInit(2, SIGUSR1, mm_TaskStopHandler);
  mm_ISRInit(3, SIGUSR2, mm_TaskContHandler);

  mm_InterruptInit();
}

int mm_ISRInit(int intno, int signo, void (*isr)())
{
  if (mm_isrNo <= MM_MAX_INT)
    mm_isrNo++;
  else
    return -1;

  mm_interruptVector[intno] = signo;
  mm_interruptRoutine[intno] = isr;

  return 0;
}

void mm_InterruptInit()
{
  int i;
  struct sigaction actISR[MM_MAX_INT];

  sigfillset(&mm_blockMask);
  sigemptyset(&mm_interruptMask);
  for (i = 0; i <= mm_isrNo; i++) {
    sigaddset(&mm_interruptMask, mm_interruptVector[i]);
    actISR[i].sa_handler = mm_interruptRoutine[i];
    sigaction(mm_interruptVector[i], &actISR[i], NULL);
  }
  sigprocmask(SIG_UNBLOCK, &mm_interruptMask, &mm_blockMask);

  mm_TASKSCHE = mm_interruptVector[1];
  mm_TASKSTOP = mm_interruptVector[2];
  mm_TASKCONT = mm_interruptVector[3];

  DebugPrint(1, "interrupt init... ok!\n");
}

void mm_CriticalSectionEnter()
{
  mm_criticalSectionFlag = TRUE;
}

void mm_CriticalSectionExit()
{
  mm_criticalSectionFlag = FALSE;
}

void mm_TaskScheHandler()
{
  if (mm_criticalSectionFlag == TRUE)
    return;
  if (mm_currTID == -1)
    mm_TimerStart();
  else
    pthread_kill(mm_currTID, mm_TASKSTOP);
  usleep(100);
  mm_currTID = mm_taskTID[mm_nextTID];
  pthread_kill(mm_currTID, mm_TASKCONT);

  DebugPrint(1, "context switching... ok!\n");
}

void mm_TaskStopHandler()
{
  DebugPrint(1, "----- catch mm_TASKSTOP\n");
  pause();
}

void mm_TaskContHandler()
{
  DebugPrint(1, "----- catch mm_TASKCONT\n");
  // nothing..
}

void mm_ContextSwitch(int next)
{
  mm_nextTID = next;
  pthread_kill(mm_mainTID, mm_TASKSCHE);
}

void *mm_TaskPrepare(void *arg)
{
  void (*fp)(void) = (void (*)())arg;
  pause();
  fp();
}

void mm_TaskGenerate(int taskID, void *func, void *arg)
{
  pthread_create(&mm_taskTID[taskID], NULL, &mm_TaskPrepare, func);
}

void mm_TimerStart()
{
  struct itimerval itimer;

  itimer.it_interval.tv_sec = 1;
  itimer.it_interval.tv_usec = 0;
  itimer.it_value = itimer.it_interval;
  usleep(1);
  setitimer(ITIMER_REAL, &itimer, NULL);

  DebugPrint(1, "timer start running... ok! %u\n", mm_currTID);
}
