/*************************************************************************** 
 * File        : pipe.h
 * Date        : 2007/04/20
 * Author      : Sun Joon Gyu
 * Modified    : 2007/04/20
 * Description : 
 * ***************************************************************************/
#ifndef _PIPE_H_
#define _PIPE_H_

#include "../types/types.h"
#include "../io/iosubsys.h"
#include "../thread/thread_op.h"

#define MAX_PIPE	255

#define NONBLOCKING	0x01
#define BLOCKING	0x00

typedef struct pipe_operation {
	int (*Create)();
	int (*Open)();
	int (*Read)();
	int (*Write)();
	int (*Close)();
	int (*Ioctl)();
	int (*Destroy)();
}pipe_op_t;


typedef struct pipe {
	unsigned short pid;
	unsigned short count;	// the number of threads using this pipe
	pipe_op_t *p_op;		// operator for pipe
	unsigned short in, out;	// index of entrance and exit of the buffer
	size_t buff_size;
	void *data;
	void *owner;		// threads which creates this pipe
} pipe_t;



char InitPipeOp();
unsigned char CreatePipe(unsigned char pid, size_t size);
pipe_t *OpenPipe(unsigned char pid, unsigned char mode);
char *ReadFromPipe(pipe_t *pipe, char *dest, size_t size);
size_t WriteToPipe(pipe_t *pipe, const char *src, size_t size );
char ClosePipe(pipe_t *pipe);
char DestroyPipe();



#endif // _PIPE_H_
