#include "thread.h"
#include "../lib/list.h"

/******							functions's prototype						******/
/******																		******/
/******																		******/


tcb_t *pRdyListHead_g;
tcb_t *pRdyListTail_g;

static stack_t *Kern_StackAlloc(fpThread_t);						// Stack Allocation
ptcb_t Kern_ThreadCreate(fpThread_t, prior_t, tid_t);		// Thread Creation
char Port_StackInit(fpThread_t, stack_t *);					// Stack Initialization

static void fun1(void *);									// Thread function for test

/******								definitions								******/
/******																		******/
/******																		******/

/*********************************************************************************
 * Description : Stack Allocation and TCB initialization
 *
 *
 *********************************************************************************/
void Kern_ThreadInit(tcb_t *newtcb,fpThread_t pT,prior_t prior,tid_t tid )
{															
  newtcb->pStack = Kern_StackAlloc(pT);						 
  newtcb->prior = prior;									
  newtcb->tid = tid;
  INIT_LIST_HEAD(&newtcb->threadList);
//  newtcb->threadList = LIST_HEAD_INIT(newtcb->threadList);
}															


/******								functions								******/
/******																		******/
/******																		******/

/*********************************************************************************
 * Description : Stack Allocation and TCB initialization
 * Arguments :
 *	fpThread_t pT	Thread pointer for allocation
 *			stack_t *pStack		Stack pointer
 * Return :
 *	char
 *		0	success
 *		1	fail
 *
 *********************************************************************************/
char Port_StackInit(fpThread_t pT, stack_t *pStack)
{
  stack_t *pStk;

  pStk = pStack;
  *pStk = (stack_t)pT;
  *(pStk--) = (stack_t)14;                // R14
  *(pStk--) = (stack_t)12;                // R12
  *(pStk--) = (stack_t)11;                // R11
  *(pStk--) = (stack_t)10;                // R10
  *(pStk--) = (stack_t)9;                // R09
  *(pStk--) = (stack_t)8;                // R08
  *(pStk--) = (stack_t)7;                // R07
  *(pStk--) = (stack_t)6;                // R06
  *(pStk--) = (stack_t)5;                // R05
  *(pStk--) = (stack_t)4;                // R04
  *(pStk--) = (stack_t)3;                // R03
  *(pStk--) = (stack_t)2;                // R02
  *(pStk--) = (stack_t)1;                // R01
  *(pStk--) = (stack_t)0;                // R00
  *(pStk--) = (stack_t)0x53;                // CPSR

  return 0;
}

static stack_t *Kern_StackAlloc(fpThread_t pT)
{
  stack_t *pStack;
  pStack = (stack_t *)MemAlloc((unsigned int *)&pStack,STACK_SIZE*sizeof(stack_t));

  Port_StackInit(pT, pStack);

  return pStack;
}


char Kern_RdyListInsert(ptcb_t newtcb)
{

	ListAddTail(&newtcb->threadList, &pRdyListHead_g->threadList);

#if 0
  if(pRdyListHead_g == NULL){
	  pRdyListHead_g = newtcb;
	  pRdyListTail_g = MemAlloc((unsigned int *)&pRdyListTail_g,sizeof(tcb_t));

	  pRdyListTail_g->pPrev = newtcb;
	  newtcb->pNext = pRdyListTail_g;
  }
  else{
	  
	  
	  pRdyListTail_g->pPrev->pNext = newtcb;
	  
	  newtcb->pPrev = pRdyListTail_g->pPrev;
	  newtcb->pNext = pRdyListTail_g;
	  pRdyListTail_g->pPrev = newtcb;

  }

#endif 
	  

  return 0;
}

ptcb_t Kern_ThreadCreate(fpThread_t pT, prior_t prior, tid_t tid)
{
  ptcb_t newtcb;
  newtcb = (ptcb_t)MemAlloc((unsigned int *)&newtcb, sizeof(tcb_t));			// TCB creates

  Kern_ThreadInit(newtcb,pT,prior,tid);									// TCB initializes and Stack allocates

  Kern_RdyListInsert(newtcb);							// Insert new TCB into Ready List

  Emul_ThreadCreate(pT,newtcb->tid);

  return newtcb;
}

// main for test
//
/*
int main()
{
  ptcb_t tcbtest;
  int i;

  tcbtest = Kern_ThreadCreate(fun1, 0, 0);
  printf("prior : %d\n", tcbtest->prior);
  printf("tid : %d\n", tcbtest->prior);

  printf("R14 : %d\n", *((stack_t *)tcbtest->pStack--));
  printf("R12 : %d\n", *((stack_t *)tcbtest->pStack--));
  for(i=0; i < 12; i++)
    printf("R%d : %d\n", 11 - i, *((stack_t *)tcbtest->pStack--));
  printf("CPSR : 0x%x\n", *((stack_t *)tcbtest->pStack--));
    
  return 0;
}

// Thread function for test
static void fun1(void *temp)
{
  printf("fun1\n");
}*/
