/* File : mm.c
 * Date : 2007/ 04/ 04
 * Written by : Sun Joon Gyu
 * Purpose : Rutines for Memory management 
*/

#include"mm.h"

unsigned int *pHeapList_g;
unsigned int *pHeapTail_g;

char heap[1024*1024];


char Kern_MemInit()
{
	
	pHeapList_g = (unsigned int *)START_HEAP_ADDR;
	pHeapTail_g = NULL;
	
	/* Initialization for a header of heap*/
	memHead_t memHeader;
	memHeader.pPrev = (pMemHead_t) pHeapList_g;
	memHeader.pNext = (pMemHead_t) pHeapTail_g;
	memHeader.status = AVAILABLE;
	memHeader.size = (unsigned int) ((END_HEAP_ADDR - pHeapList_g ) / BLOCK_SIZE) - BLOCK_SIZE; /* should subtract the size of one memory block */

	if(memHeader.size <= 0)
		return FAILED;

	/*Put the header just finished initializing into first location of heap*/
	*((pMemHead_t)pHeapList_g) = memHeader;

	return SUCCESS;

}

pMemHead_t Kern_WantedBlkFind(size_t size)
{
	pMemHead_t pBlk = (pMemHead_t)pHeapList_g;

	for(;pBlk != NULL;pBlk = pBlk->pNext){
		if(pBlk->size >= size){
			return pBlk;
		}
	}

	return NULL;
		
}

char Kern_MemBlkGet(pMemHead_t pBlkAddr,size_t size)
{

	size += (BLOCK_SIZE-1);
	size /= BLOCK_SIZE;
	/* Just in case there is something wrong with memory block pointer */
	if((pBlkAddr->size) == size){
		pBlkAddr->status = ALLOCATED;
		pBlkAddr->pPrev->pNext = pBlkAddr->pNext;
		pBlkAddr->pNext->pPrev = pBlkAddr->pPrev;
	}
	else if((pBlkAddr->size) > size){
		pBlkAddr->status = ALLOCATED;

		memHead_t memHeader;
		size_t extraBlkSize = pBlkAddr->size - size - 1;   /* should subtract an additional memory control header */
		memHeader.size = extraBlkSize;
		memHeader.pPrev = pBlkAddr->pPrev;
		memHeader.pNext = pBlkAddr->pNext;
		memHeader.status = AVAILABLE;

		
		pMemHead_t pExtraMemBlk = (pMemHead_t)((char *)pBlkAddr + ((size + 1) * BLOCK_SIZE)); 
		pHeapList_g = (unsigned int *)pExtraMemBlk;

		*((pMemHead_t)pExtraMemBlk) = memHeader;

		pBlkAddr->pPrev->pNext = pExtraMemBlk;
		pBlkAddr->pNext->pPrev = pExtraMemBlk;
		pBlkAddr->pNext = pExtraMemBlk;
		pBlkAddr->size = size;
		pBlkAddr->status = ALLOCATED;
	}
	else{
		return FAILED;
	}

	return SUCCESS;	

}	
	
char Kern_BlkTblRegister(unsigned int *pReqVar, pMemHead_t pBlkAddr)
{

	//pBlkAddr->pAllocNext = pCurrThread->memList->pAllocNext;
	//pCurrThread->memList = pBlkAddr;
	pBlkAddr->pVar = pReqVar;
	
}


char MemRelease(void *p)
{
	pMemHead_t pIter;

	for(pIter = (pMemHead_t)((char *)p - BLOCK_SIZE); pIter != NULL; pIter = pIter->pAllocNext){
		if(pIter == p){
			if(pIter->pPrev != NULL)
				pIter->pAllocPrev->pAllocNext = pIter->pAllocNext;
			if(pIter->pNext != NULL)
				pIter->pAllocNext->pAllocPrev = pIter->pAllocPrev;
		}
	}


	if(((pMemHead_t)((char *)p - BLOCK_SIZE))->pPrev->status != ALLOCATED){
		/* merging */
	}
	if(((pMemHead_t)((char *)p - BLOCK_SIZE))->pNext->status != ALLOCATED){
		/* merging */
	}
	
			
			
		
	
}

unsigned int *MemAlloc(unsigned int *pReqVar, size_t size)
{

	unsigned char err;
	pMemHead_t pBlkAddr;

	if( (pBlkAddr = Kern_WantedBlkFind(size)) == NULL){
		/* The thread tried to get memory from heap area should be blocked!! */
		/* DbgPrint("Not enough memory in Heap area\n"); */
	}

	if(Kern_MemBlkGet(pBlkAddr, size) == FAILED){
		/* DbgPrint("Something wrong!!\n"); */
	}

	if((Kern_BlkTblRegister(pReqVar, pBlkAddr)) ){
		/* The block requested to regist on block management table for compaction is already allocated!! */
		/* DbgPrint("The block requested is already allocated before!!\n"); */
	}

	return (unsigned int *)((char *)pBlkAddr + BLOCK_SIZE);	/* return the address except memory header */
}
int main()
{
	int k = Kern_MemInit();
	char *p;
	unsigned int *addr;
	int i;
	addr = pHeapList_g;
	printf("%d",k);
	printf("%p", ((unsigned int *)addr));
	p = MemAlloc((unsigned int *)&p,61);
	printf("%p\n",p);	
	for(i = 0; i < 50; i++){
		p[i] = 'i';
	}
	p[i] = '\0';
	printf("%s",p);	
		
	
	printf("%p", ((pMemHead_t)addr)->pPrev);
}
